"""This module allows registration and management of jobruns, jobrun actions and log files"""

try:
    from urllib.parse import urlparse
except ImportError:
    import urlparse
import getpass
import json
import requests
import shapely
import dateutil.parser

CALLBACK_BASE_URL = 'http://jobcontroltest.vgt.vito.be:8090'

REST_HEADERS = {'Accept': 'application/json',
                'Content-Type': 'application/json'}

class CallbackClient(object):
    """Client allowing to register jobruns, jobrun actions and logs."""

    def __init__(self, baseurl=CALLBACK_BASE_URL):

        self.baseurl = baseurl

    def register_jobrun(self, job_id, username=getpass.getuser(), start_time=None):
        """Register a jobrun"""

        data = {"job_id": job_id,
                "username": username}
        if start_time is not None:
            data["start_time"] = start_time.isoformat()

        url = urlparse.urljoin(self.baseurl, 'jobruns/')
        response = requests.post(url,
                                 json=data,
                                 headers=REST_HEADERS)

        if response.status_code == requests.codes.created:
            return JobRun.build(response.json())
        else:
            response.raise_for_status()

    def jobrun_finished(self, jobrun_id, time=None):
        """Change the state of a jobrun to FINISHED"""
        return self._change_jobrun_state(jobrun_id, "FINISHED", time)

    def jobrun_failed(self, jobrun_id, time=None):
        """Change the state of a jobrun to FAILED"""
        return self._change_jobrun_state(jobrun_id, "FAILED", time)

    def jobrun_action_finished(self, jobrun_id, jobrun_action_id, time=None):
        """Change the state of a jobrun action to FINISHED"""
        return self._change_jobrun_action_state(jobrun_id, jobrun_action_id, "FINISHED", time)

    def jobrun_action_failed(self, jobrun_id, jobrun_action_id, time=None):
        """Change the state of a jobrun action to FAILED"""
        return self._change_jobrun_action_state(jobrun_id, jobrun_action_id, "FAILED", time)

    def jobrun_action_product_running(self, jobrun_id, jobrun_action_id, input_product):
        """Change the state of a jobrun action input product to RUNNING"""
        return self._change_jobrun_action_product_state(jobrun_id, jobrun_action_id, input_product,
                                                        "RUNNING")

    def jobrun_action_product_finished(self, jobrun_id, jobrun_action_id, input_product):
        """Change the state of a jobrun action input product to RUNNING"""
        return self._change_jobrun_action_product_state(jobrun_id, jobrun_action_id, input_product,
                                                        "FINISHED")

    def jobrun_action_product_failed(self, jobrun_id, jobrun_action_id, input_product):
        """Change the state of a jobrun action input product to RUNNING"""
        return self._change_jobrun_action_product_state(jobrun_id, jobrun_action_id, input_product,
                                                        "FAILED")

    def register_jobrun_action(self, jobrun_id, name, start_time=None, input_products=None):
        """Register a jobrun action"""

        data = {"name": name}
        if start_time is not None:
            data["start_time"] = start_time.isoformat()
        if input_products is not None:
            data["input_products"] = input_products

        data_json = json.dumps(data, cls=InputProductEncoder)

        url = self.baseurl + '/jobruns/' + jobrun_id + '/actions/'
        response = requests.post(url,
                                 data=data_json,
                                 headers=REST_HEADERS)

        if response.status_code == requests.codes.created:
            return JobRunAction.build(response.json())
        else:
            response.raise_for_status()

    def add_jobrun_action_log(self, jobrun_id, jobrun_action_id, message):
        """Add a log entry for a jobrun action"""

        self._add_jobrun_action_log(jobrun_id, jobrun_action_id, None, message)

    def add_jobrun_action_product_log(self, jobrun_id, jobrun_action_id,
                                      input_product_ref, message):
        """Add a log entry for a jobrun action on a specific input product"""

        self._add_jobrun_action_log(jobrun_id, jobrun_action_id, input_product_ref, message)

    def _change_jobrun_state(self, jobrun_id, state, time=None):

        data = {"state": state}
        if time is not None:
            data["time"] = time.isoformat()

        url = self.baseurl + '/jobruns/' + jobrun_id
        requests.patch(url, json=data, headers=REST_HEADERS)

    def _change_jobrun_action_state(self, jobrun_id, jobrun_action_id, state, time=None):

        data = {"state": state}
        if time is not None:
            data["time"] = time.isoformat()

        url = self.baseurl + '/jobruns/' + jobrun_id + '/actions/' + jobrun_action_id
        requests.patch(url, json=data, headers=REST_HEADERS)

    def _change_jobrun_action_product_state(self, jobrun_id, jobrun_action_id,
                                            input_product_ref, state):

        data = {"state": state,
                "input_product_reference": input_product_ref}

        url = self.baseurl + '/jobruns/' + jobrun_id + '/actions/' + jobrun_action_id + '/products/'
        requests.patch(url, json=data, headers=REST_HEADERS)

    def _add_jobrun_action_log(self, jobrun_id, jobrun_action_id, input_product_ref, message):

        data = {"message": message}
        if input_product_ref is not None:
            data["input_product_reference"] = input_product_ref

        requests.post(self.baseurl
                      + '/jobruns/' + jobrun_id + '/actions/'
                      + jobrun_action_id + '/logs',
                      json=data,
                      headers=REST_HEADERS)

class JobRun(object):
    """This class represents a jobrun."""

    def __init__(self, id_, job_id, username, state, start_time, end_time=None):
        self.id_ = id_
        self.job_id = job_id
        self.username = username
        self.state = state
        self.start_time = start_time
        self.end_time = end_time

    def __str__(self):
        return "{0} {1}".format(self.job_id, self.username)

    @staticmethod
    def build(json_input):
        """Build a JobRun object from JSON input."""
        end_time = None
        if json_input['end_time'] is not None:
            end_time = dateutil.parser.parse(json_input['end_time'])
        return JobRun(json_input['id'],
                      json_input['job_id'],
                      json_input['username'],
                      json_input['state'],
                      dateutil.parser.parse(json_input['start_time']),
                      end_time)


class JobRunAction(object):
    """This class represents a jobrun action."""

    def __init__(self, id_, jobrun_id, name, state, start_time, end_time=None):
        self.id_ = id_
        self.jobrun_id = jobrun_id
        self.name = name
        self.state = state
        self.start_time = start_time
        self.end_time = end_time
        self.input_products = []

    def __str__(self):
        return "{0} {1} {2}".format(self.id_, self.jobrun_id, self.name)

    @staticmethod
    def build(json_input):
        """Build a JobRunAction object from JSON input."""
        end_time = None
        if json_input['end_time'] is not None:
            end_time = dateutil.parser.parse(json_input['end_time'])
        return JobRunAction(json_input['id'],
                            json_input['jobrun_id'],
                            json_input['name'],
                            json_input['state'],
                            dateutil.parser.parse(json_input['start_time']),
                            end_time)

    def add_input_product(self, input_product):
        """Add input product to list of action input products."""
        self.input_products += input_product


class InputProduct(object):
    """This class represents a jobrun action input product."""

    def __init__(self, reference, timestamp=None, geometry=None):
        self.reference = reference
        self.timestamp = timestamp
        self.geometry = geometry

    def __str__(self):
        return "{0}".format(self.reference)


class InputProductEncoder(json.JSONEncoder):
    """This class JSON encodes an InputProduct."""
    def default(self, o): # pylint: disable=E0202
        if isinstance(o, InputProduct):
            dic = {'reference': o.reference}
            if o.timestamp is not None:
                dic['timestamp'] = o.timestamp.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            if o.geometry is not None:
                dic['geometry'] = o.geometry.__geo_interface__
            return dic
        elif isinstance(o, shapely.geometry.base.BaseGeometry):
            return o.__geo_interface__
        else:
            return json.JSONEncoder.default(self, o)
